#include "../Framework/Framework.h"

#include "Poly.h"

class MainApp : public OpenGLApp {
protected:
	TextureID brickBump, floorBump, wallBump, lightTexture;

	TextureID vectorField, renderTexture, ditherMap;

	ShaderID vertexShader;
	Set <Poly> polys;
	Set <Light> lights;

	void drawWorld();
public:
	void selectPixelFormat(PixelFormat &pixelFormat);

	bool load();
	bool unload();
	bool drawFrame();
	
	void resetCamera();
};

void MainApp::resetCamera(){
	position = Vertex(350,200,-350);
	Wx = PI / 5;
	Wy = PI / 4;
	Wz = 0;
}

void MainApp::selectPixelFormat(PixelFormat &pixelFormat){
	pixelFormat.redBits   = 8;
	pixelFormat.greenBits = 8;
	pixelFormat.blueBits  = 8;
	pixelFormat.alphaBits = 0;

	pixelFormat.depthBits = 16;
	pixelFormat.stencilBits = 0;
}


bool MainApp::load(){

	String supported, missing;
	if (GL_ATI_fragment_shader || GL_NV_register_combiners){
		if (GL_ATI_fragment_shader)   supported += "GL_ATI_fragment_shader\n";
		if (GL_NV_register_combiners) supported += "GL_NV_register_combiners\n";
	} else {
		missing += "GL_ATI_fragment_shader\nGL_NV_register_combiners\n";
	}
    //if (WGL_ARB_render_texture)    supported += "WGL_ARB_render_texture\n";    else missing += "WGL_ARB_render_texture\n";
	if (GL_ARB_texture_env_dot3)   supported += "GL_ARB_texture_env_dot3\n";   else missing += "GL_ARB_texture_env_dot3\n";
    if (GL_EXT_texture3D)          supported += "GL_EXT_texture3D\n";          else missing += "GL_EXT_texture3D\n";
    if (GL_EXT_texture_edge_clamp) supported += "GL_EXT_texture_edge_clamp\n"; else missing += "GL_EXT_texture_edge_clamp\n";

    if (missing.getLength() > 0){
        ALERT("The following required OpenGL extensions are supported:\n\n" + supported +
			  "\nToo bad, these required OpenGL extensions aren't supported by the graphic card and driver you're using:\n\n" + missing +
              "\nThis application will now exit.", "Missing extensions");
        return false;
    }

	renderTexture = renderer->addRenderTexture(256, 256, false, false);

	vectorField = renderer->addVectorField(64, TEX_NOMIPMAPPING);

	brickBump = renderer->addTexture("../Textures/HumusBrickBump.png",  TEX_NORMALMAP | TEX_USE_16BIT);
	floorBump = renderer->addTexture("../Textures/brickfloorBump.png",  TEX_NORMALMAP | TEX_USE_16BIT);
	wallBump  = renderer->addTexture("../Textures/wallPaper20Bump.png", TEX_NORMALMAP | TEX_USE_16BIT);

	lightTexture = renderer->addParticleTexture(64);
	textTexture  = renderer->addTexture("../Textures/font2.dds");

	int ditherFactor = 75;
	int ditherFactorHalf = ditherFactor / 2;
	unsigned char *lookup = new unsigned char[512 * 512];
	for (int i = 0; i < 512 * 512; i++){
		 lookup[i] = (rand() % ditherFactor) + 128 - ditherFactorHalf;
	}
	Image *img = new Image();
	img->loadFromMemory(lookup, 512, 512, FORMAT_I8, true, false);
	ditherMap = renderer->addTexture(img, TEX_NOLINEARFILTER);

	if (GL_ATI_fragment_shader){
		GLuint shaderID = glGenFragmentShadersATI(1);
		glBindFragmentShaderATI(shaderID);
   		glBeginFragmentShaderATI();
			GLfloat gray[] = {0.30f, 0.59f, 0.11f, 0.0f};
			glSetFragmentShaderConstantATI(GL_CON_0_ATI, gray);

			glSampleMapATI(GL_REG_0_ATI, GL_TEXTURE0_ARB, GL_SWIZZLE_STR_ATI);
			glSampleMapATI(GL_REG_1_ATI, GL_TEXTURE1_ARB, GL_SWIZZLE_STR_ATI);

			glColorFragmentOp2ATI(GL_DOT3_ATI, GL_REG_0_ATI, GL_NONE, GL_NONE,
											   GL_REG_0_ATI, GL_NONE, GL_NONE,
											   GL_CON_0_ATI, GL_NONE, GL_NONE);

			glColorFragmentOp2ATI(GL_ADD_ATI,  GL_REG_0_ATI, GL_NONE, GL_NONE,
											   GL_REG_0_ATI, GL_NONE, GL_NONE,
											   GL_REG_1_ATI, GL_NONE, GL_BIAS_BIT_ATI);

			glColorFragmentOp3ATI(GL_CND_ATI,  GL_REG_0_ATI, GL_NONE, GL_NONE,
											   GL_ONE,       GL_NONE, GL_NONE,
											   GL_ZERO,      GL_NONE, GL_NONE,
											   GL_REG_0_ATI, GL_NONE, GL_NONE);

		glEndFragmentShaderATI();
		renderer->addFragmentShader(GL_FRAGMENT_SHADER_ATI, shaderID);
	} else {
        GLint nGeneralCombiners = 0;
        glGetIntegerv(GL_MAX_GENERAL_COMBINERS_NV, &nGeneralCombiners);
		
        if (nGeneralCombiners < 3){
			ALERT("Too few general combiners", "Error");
			return false;
        } else {

			GLuint shaderID = glGenLists(1);
			glNewList(shaderID, GL_COMPILE);
				
				glCombinerParameteriNV(GL_NUM_GENERAL_COMBINERS_NV, 3);

				GLfloat gray[] = {0.30f, 0.59f, 0.11f, 0.0f};
				glCombinerParameterfvNV(GL_CONSTANT_COLOR0_NV, gray);
				glCombinerInputNV(GL_COMBINER0_NV,  GL_RGB, GL_VARIABLE_A_NV, GL_TEXTURE0_ARB,       GL_UNSIGNED_IDENTITY_NV, GL_RGB);
				glCombinerInputNV(GL_COMBINER0_NV,  GL_RGB, GL_VARIABLE_B_NV, GL_CONSTANT_COLOR0_NV, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
				glCombinerOutputNV(GL_COMBINER0_NV, GL_RGB, GL_TEXTURE0_ARB, GL_DISCARD_NV, GL_DISCARD_NV, GL_NONE, GL_NONE, GL_TRUE, GL_FALSE, GL_FALSE);

				glCombinerInputNV(GL_COMBINER1_NV,  GL_ALPHA, GL_VARIABLE_A_NV, GL_TEXTURE0_ARB, GL_SIGNED_IDENTITY_NV,  GL_BLUE);
				glCombinerInputNV(GL_COMBINER1_NV,  GL_ALPHA, GL_VARIABLE_B_NV, GL_ZERO,         GL_UNSIGNED_INVERT_NV,  GL_BLUE);
				glCombinerInputNV(GL_COMBINER1_NV,  GL_ALPHA, GL_VARIABLE_C_NV, GL_TEXTURE1_ARB, GL_HALF_BIAS_NORMAL_NV, GL_BLUE);
				glCombinerInputNV(GL_COMBINER1_NV,  GL_ALPHA, GL_VARIABLE_D_NV, GL_ZERO,         GL_UNSIGNED_INVERT_NV,  GL_BLUE);
				glCombinerOutputNV(GL_COMBINER1_NV, GL_ALPHA, GL_DISCARD_NV, GL_DISCARD_NV, GL_SPARE0_NV, GL_NONE, GL_NONE, GL_FALSE, GL_FALSE, GL_FALSE);

				glCombinerInputNV(GL_COMBINER2_NV,  GL_RGB, GL_VARIABLE_A_NV, GL_ZERO, GL_SIGNED_IDENTITY_NV, GL_RGB);
				glCombinerInputNV(GL_COMBINER2_NV,  GL_RGB, GL_VARIABLE_B_NV, GL_ZERO, GL_SIGNED_IDENTITY_NV, GL_RGB);
				glCombinerInputNV(GL_COMBINER2_NV,  GL_RGB, GL_VARIABLE_C_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);
				glCombinerInputNV(GL_COMBINER2_NV,  GL_RGB, GL_VARIABLE_D_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);
				glCombinerOutputNV(GL_COMBINER2_NV, GL_RGB, GL_DISCARD_NV, GL_DISCARD_NV, GL_SPARE0_NV,   GL_NONE, GL_NONE, GL_FALSE, GL_FALSE, GL_TRUE);

				glFinalCombinerInputNV(GL_VARIABLE_A_NV, GL_ZERO,      GL_UNSIGNED_IDENTITY_NV, GL_RGB);
				glFinalCombinerInputNV(GL_VARIABLE_B_NV, GL_ZERO,      GL_UNSIGNED_IDENTITY_NV, GL_RGB);
				glFinalCombinerInputNV(GL_VARIABLE_C_NV, GL_ZERO,      GL_UNSIGNED_IDENTITY_NV, GL_RGB);
				glFinalCombinerInputNV(GL_VARIABLE_D_NV, GL_SPARE0_NV, GL_UNSIGNED_IDENTITY_NV, GL_RGB);

    		glEndList();

			renderer->addFragmentShader(GL_REGISTER_COMBINERS_NV, shaderID);
        }
    }

 


	projection.loadProjectionMatrix(90 * PI / 180, 0.75f, 0.1f, 10000);

	polys.add(Poly(Vertex(-256, -100,  256), Vertex( 256, -100,  256), Vertex(-256, -100, -256), brickBump, 1, 1));
	polys.add(Poly(Vertex(-256, -100, -256), Vertex( 256, -100, -256), Vertex(-256, -256, -256), brickBump, 1, 0.25f));
	polys.add(Poly(Vertex( 256, -100, -256), Vertex( 256, -100,  256), Vertex( 256, -256, -256), brickBump, 1, 0.25f));
	polys.add(Poly(Vertex( 256, -100,  256), Vertex(-256, -100,  256), Vertex( 256, -256,  256), brickBump, 1, 0.25f));
	polys.add(Poly(Vertex(-256, -100,  256), Vertex(-256, -100, -256), Vertex(-256, -256,  256), brickBump, 1, 0.25f));

	polys.add(Poly(Vertex(-70,  70, -70), Vertex( 70,  70, -70), Vertex(-70, -70, -70), floorBump, 1, 1));
	polys.add(Poly(Vertex( 70,  70, -70), Vertex( 70,  70,  70), Vertex( 70, -70, -70), floorBump, 1, 1));
	polys.add(Poly(Vertex( 70,  70,  70), Vertex(-70,  70,  70), Vertex( 70, -70,  70), floorBump, 1, 1));
	polys.add(Poly(Vertex(-70,  70,  70), Vertex(-70,  70, -70), Vertex(-70, -70,  70), floorBump, 1, 1));
	polys.add(Poly(Vertex(-70,  70,  70), Vertex( 70,  70,  70), Vertex(-70,  70, -70), floorBump, 1, 1));
	polys.add(Poly(Vertex(-70, -70, -70), Vertex( 70, -70, -70), Vertex(-70, -70,  70), floorBump, 1, 1));

	
	polys.add(Poly(Vertex(-512, -256,  512), Vertex( 512, -256,  512), Vertex(-512, -256,  256), floorBump, 4, 1));
	polys.add(Poly(Vertex(-512, -256, -256), Vertex( 512, -256, -256), Vertex(-512, -256, -512), floorBump, 4, 1));
	polys.add(Poly(Vertex(-512, -256,  256), Vertex(-256, -256,  256), Vertex(-512, -256, -256), floorBump, 1, 2));
	polys.add(Poly(Vertex( 256, -256,  256), Vertex( 512, -256,  256), Vertex( 256, -256, -256), floorBump, 1, 2));

	polys.add(Poly(Vertex(-512,  512,  512), Vertex( 512,  512,  512), Vertex(-512, -256,  512), wallBump, 3, 3));
	polys.add(Poly(Vertex(-512,  512, -512), Vertex( 512,  512, -512), Vertex(-512,  512,  512), wallBump, 3, 3));
	polys.add(Poly(Vertex( 512,  512,  512), Vertex( 512,  512, -512), Vertex( 512, -256,  512), wallBump, 3, 3));
	polys.add(Poly(Vertex( 512,  512, -512), Vertex(-512,  512, -512), Vertex( 512, -256, -512), wallBump, 3, 3));
	polys.add(Poly(Vertex(-512,  512, -512), Vertex(-512,  512,  512), Vertex(-512, -256, -512), wallBump, 3, 3));

	lights.add(Light(Vertex(0,0,0), Color(1,1,1), 800));
	lights.add(Light(Vertex(0,0,0), Color(1,1,1), 800));
	return true;
}

bool MainApp::unload(){
	polys.clear();
	lights.clear();

	return true;
}

void MainApp::drawWorld(){
	glEnable(GL_CULL_FACE);
	glCullFace(GL_FRONT);
	
	for (unsigned int j = 0; j < lights.getSize(); j++){
		for (unsigned int i = 0; i < polys.getSize(); i++){
			polys[i].drawLit(renderer, vectorField, lights[j], j > 0);
		}	
	}

	glDisable(GL_CULL_FACE);
}


bool MainApp::drawFrame(){

	Vertex vert(15 * sinf(currTime * 2.0f), 15 * sinf(currTime * 3.24f) + 14, 35 * cosf(currTime * 0.73f));
	vert.normalize();
	vert *= 100 * (sinf(currTime + 0.43f) + 3.0f);
	lights[0].setPosition(vert);

	Vertex vert2(25 * cosf(currTime * 1.0f) * sinf(currTime * 1.3f), 15 * sinf(currTime * 3.24f) + 18, 20 * sinf(currTime * 1.0f) * cosf(currTime * 0.72f));

	vert2.normalize();
	vert2 *= 70 * (cosf(currTime) + 3.0f);

	lights[1].setPosition(vert2);


	renderer->changeRenderTarget(renderTexture);
	
		glMatrixMode(GL_PROJECTION);
		//glPushMatrix();

		glLoadMatrixf(projection);
		
		modelView.loadRotateZXY(-Wx, -Wy, -Wz);
		modelView.translate(-position);
		
		glMatrixMode(GL_MODELVIEW);
		//glPushMatrix();
		glLoadMatrixf(modelView);
		
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		
		glDepthFunc(GL_LEQUAL);
		glEnable(GL_DEPTH_TEST);

		drawWorld();

		renderer->setBlending(GL_ONE, GL_ONE);
		renderer->setTextures(lightTexture);
		renderer->apply();

		glDepthMask(GL_FALSE);

		Vertex dx(modelView[0], modelView[4], modelView[8]);
		Vertex dy(modelView[1], modelView[5], modelView[9]);
		Vertex v,r;
		for (unsigned int i = 0; i < lights.getSize(); i++){
			v = lights[i].getPosition();
			Color col = lights[i].getColor();
			glColor4fv(col);
			
			glBegin(GL_QUADS);
				r = v - 30 * dx - 30 * dy;
				glTexCoord2f(0,0);
				glVertex3fv((GLfloat *) &r);
			
				r = v + 30 * dx - 30 * dy;
				glTexCoord2f(1,0);
				glVertex3fv((GLfloat *) &r);
			
				r = v + 30 * dx + 30 * dy;
				glTexCoord2f(1,1);
				glVertex3fv((GLfloat *) &r);
			
				r = v - 30 * dx + 30 * dy;
				glTexCoord2f(0,1);
				glVertex3fv((GLfloat *) &r);
			glEnd();
		}
		glDepthMask(GL_TRUE);

		
/*		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		*/
	renderer->changeRenderTarget();


	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0,1,0,1,1,-1);

	renderer->setTextures(renderTexture, ditherMap);
	renderer->setFragmentShader(0);
	renderer->apply();

	glBegin(GL_QUADS);
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0, 0);
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0, 0);
		glVertex2f(0,0);

		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1, 0);
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1, 0);
		glVertex2f(1,0);

		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1, 1);
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1, 1);
		glVertex2f(1,1);

		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0, 1);
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0, 1);
		glVertex2f(0,1);
	glEnd();

	return true;
}

DECLARE_MAIN_APP(MainApp)
